#!/usr/bin/python
import re
import sys
import signal
import os
import math
import subprocess
import threading
import time
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()
WORKTAG = 0
ANSWERTAG = 1
DIETAG = 2

if (len(sys.argv) < 2):
  print "give hg id as argument"
  sys.exit(-1)

hgid = sys.argv[1]

initials = "INITIALS"
initialfile = hgid + "/" + initials
rawdir = hgid + "/raw"
returndir = hgid + "/return"

def tuple5_of_list(l):
  [a, b, c, d, e] = l[:5]
  return (a, b, c, d, e)

def read_initials_file(filename):
  with open(filename, 'r') as infile:
    return {tuple5_of_list(line.split()):line.split()[5] for line in infile.readlines()}

def write_initials_file(entries, filename):
  with open(filename, 'w') as outfile:
    for key in entries:
      outfile.write (" ".join(key) + " " + entries[key] + "\n")

def first_new_entry(entries):
  for key in entries:
    # Only start from the section of interest...
    if (int(key[4]) == 27 * (2**int(key[0]))):
      if (entries[key] == "NEW"):
        return key

def push_entry(entries, key, new_val):
  if key in entries:
    if entries[key] == "NEW" or not (new_val == "NEW"):
      print ("overwriting " + entries[key] + " with " + new_val)
      entries[key] = new_val
  else:
    entries[key] = new_val

def get_next_input(entries,server):
  return_entry = None
  entry = first_new_entry(entries)
  if not entry is None:
    entries[entry] = "RUNNING_on_" + str(server)
  return entry

def put_inputs(entries, new_entries):
  for entry in new_entries:
    push_entry(entries, entry, new_entries[entry])

def filename_of_entry(entry):
  return "R_" + entry[0] + "_" + entry[1] + "_" + entry[2] + "_" + entry[3] + "_" + entry[4] + ".log"

def run_lorenz(entry):
  new_entries = {entry:"FAILED"}
  try:
    rawfile = rawdir+"/"+filename_of_entry(entry)
    returnfile = returndir+"/"+filename_of_entry(entry)
    command = "./" + hgid + "/lorenz_mpi_wrapper"
    args = [hgid, rawfile, returnfile, entry[0] ,entry[1], entry[2], entry[3]]
    print ("Worker Spawning " + str(entry))
    icomm = MPI.COMM_SELF.Spawn(command, args)
    print ("Worker Barrier " + str(entry))
    icomm.Barrier()
    print ("Worker Received " + str(entry))
    icomm.Disconnect()
    #TODO: bad: racing on FS?!
    time.sleep(1)

    with open(rawfile, "r") as infile:
      with open(returnfile, "w") as outfile:
        returning = False
        for line in infile:
          if returning and not (line.startswith("=")):
            # TODO: adapt p and w values?!
            outfile.write(entry[0] + " " + entry[1] + " " + line)
          if "= Initial Rectangles of Return:" in line:
            returning = True
    with open(returnfile, "r") as infile:
      new_marker = "NEW"
      new_entries = {tuple5_of_list(line.split()):new_marker for line in infile}
      if new_entries == {}:
        print "Empty file"
        new_entries[entry] = "FAILED"
      else:
        new_entries[entry] = "RETURNED"
  except :
    print "=========== Caught some exception =============="
    pass
  return new_entries

def get_next_worker(workers):
  if workers:
    return workers.pop(0)

# Master
if rank == 0:
  if not os.path.exists(rawdir):
    os.makedirs(rawdir)
  if not os.path.exists(returndir):
    os.makedirs(returndir)

  status = MPI.Status()
  entries = read_initials_file(initialfile)
  waiting = range(1, size)

  while 1:
    try:
      num_waiting = len(waiting)
      worker = get_next_worker(waiting)
      entry = None
      if worker:
        entry = get_next_input(entries, worker)
      if not entry or not worker:
        if num_waiting == size - 1:
          print "No entries and all workers finished --> break"
          break
        else:
          if worker:
            waiting.append(worker) # previously popped worker
          if entry:
            put_inputs([entry], entries)
          print (str (len(entries)) + " entries " + " and " + str(num_waiting) + " waiting, but working processors: receiving")
          new_entries = comm.recv(obj=None, source=MPI.ANY_SOURCE, tag=WORKTAG, status=status)
          worker = status.Get_source()
          put_inputs(entries, new_entries)
          waiting.append(worker)
      else:
        print "Starting worker " + str(worker) + " on " + str(entry)
        comm.send(obj=entry, dest=worker, tag=WORKTAG)

      write_initials_file(entries, initialfile)
    except KeyboardInterrupt:
      break

  print "Killing Slaves..."

  for i in range(1,size):
      comm.send(obj=None, dest=i, tag=DIETAG)

  print "Killed Slaves"

# Slaves
else:
  status = MPI.Status()
  while 1:
    entry = comm.recv(obj=None, source=0, tag=MPI.ANY_TAG, status=status)
    if status.Get_tag() == DIETAG: break
    comm.send(obj=run_lorenz(entry), dest=0)
