#!/usr/bin/python
import re
import sys
import signal
import os
import math
import subprocess
import threading
import time

if (len(sys.argv) < 3):
  print "give hg id and remote/local as arguments"
  sys.exit(-1)

hgid = sys.argv[1]
remotelocal = sys.argv[2]
servers = []
path = ""
if (remotelocal == "remote"):
#  servers = ["macbroy23", "macbroy24", "macbroy25", "macbroy26", "macbroy27", "macbroy28", "macbroy23", "macbroy24", "macbroy25", "macbroy26", "macbroy27", "macbroy28"]
  servers = ["macbroy23","macbroy24","macbroy25","macbroy26","macbroy23","macbroy24","macbroy25","macbroy26"]
  path = "/home/immler/work/afp/afp-private/thys/Ordinary_Differential_Equations/Ex/Lorenz/"
elif (remotelocal == "local"):
  servers = ["localhost","localhost","localhost"]
  path = "/Users/immler/work/afp/afp-private/thys/Ordinary_Differential_Equations/Ex/Lorenz"
else:
  print "Give remote or local as argument"
  sys.exit(-1)
cmd = "./lorenz.py"
initials = "INITIALS"
initialfile = hgid + "/" + initials

def tuple4_of_list(l):
  [a, b, c, d] = l[:4]
  return (a, b, c, d)

def read_initials_file(filename):
  with open(filename, 'r') as infile:
    return {tuple4_of_list(line.split()):line.split()[4] for line in infile.readlines()}

def write_initials_file(filename):
  with open(filename, 'w') as outfile:
    for key in LorenzThread.entries:
      outfile.write (" ".join(key) + " " + LorenzThread.entries[key] + "\n")

def first_new_entry():
  for key in LorenzThread.entries:
    if (LorenzThread.entries[key] == "NEW" or LorenzThread.entries[key] == "EXISTING"):
      return key

def push_entry(key, new_val):
  if key in LorenzThread.entries:
    if LorenzThread.entries[key] == "NEW" or LorenzThread.entries[key] == "EXISTING" or not (new_val == "NEW" or new_val == "EXISTING"):
      print ("overwriting " + LorenzThread.entries[key] + " with " + new_val)
      LorenzThread.entries[key] = new_val
  else:
    LorenzThread.entries[key] = new_val

def get_next_input(server):
  LorenzThread.entrieslock.acquire()
  return_entry = None
  entry = first_new_entry()
  if not entry is None:
    if LorenzThread.entries[entry] == "EXISTING":
      LorenzThread.entries[entry] = "FAKING_on_" + server
    else:
      LorenzThread.entries[entry] = "RUNNING_on_" + server
  LorenzThread.entrieslock.release()
  return entry

def put_inputs(new_entries):
  LorenzThread.entrieslock.acquire()
  for entry in new_entries:
    push_entry(entry, new_entries[entry])
  LorenzThread.entrieslock.release()

def filename_of_entry(entry):
  return "R_" + entry[0] + "_" + entry[1] + "_" + entry[2] + "_" + entry[3] + ".log"

def remote_lorenz(server, entry):
  remote = ['\"cd', path, ";", "nice", cmd, hgid] + list(entry) + ['\"']
  print "Calling"
  call = " ".join(["ssh", server, " ".join(remote)])
  # call = " ".join([cmd, hgid] + entry[:4])
  print call
  if "FAKING_on_" in LorenzThread.entries[entry]:
    print "Faking call, expecting return file to be there"
  else:
    subprocess.call(call, shell=True)
  returnfile = hgid + "/return/" + filename_of_entry(entry)
  print "Checking"
  print returnfile
  if os.path.exists(returnfile):
    with open(returnfile, 'r') as infile:
      new_marker = "NEW"
      if "FAKING_on_" in LorenzThread.entries[entry]:
        new_marker = "EXISTING"
      new_entries = {tuple4_of_list(line.split()):new_marker for line in infile}
      if new_entries == {}:
        print "Empty file"
        new_entries[entry] = "FAILED"
      else:
        new_entries[entry] = "RETURNED"
      print "returned:"
      print new_entries
      put_inputs(new_entries)
  else:
    print "Failed, no such file"
    put_inputs({entry:"FAILED"})

class LorenzThread (threading.Thread):
  stop_threads = False
  entrieslock = threading.Lock()
  entries = {}

  def __init__(self, threadID, server):
    threading.Thread.__init__(self)
    self.threadID = threadID
    self.name = server
    self.server = server
  def run(self):
    try:
      print "Starting " + self.name
      while not LorenzThread.stop_threads:
        entry = None
        while entry is None and not LorenzThread.stop_threads:
          entry = get_next_input(self.server)
  #        if entry is None:
  #          print "No entry for " + self.name
          time.sleep(0.1)
        if not entry is None and not LorenzThread.stop_threads:
          print (self.name + "working on: " + " ".join(entry))
          remote_lorenz(self.server, entry)
    except KeyboardInterrupt:
      print "Interrupt in Thread"
    print "Killing processes:"
    call = " ".join(["ssh", self.server, "killall", "lorenzi"])
    print call
    subprocess.call(call, shell=True)
    print "Exiting " + self.name

def main():
  threads = []
  print "Reading "
  print initialfile
  LorenzThread.entries = read_initials_file(initialfile)
  print id(LorenzThread.entries)
  print id(LorenzThread.entrieslock)
  for server in servers:
    thread = LorenzThread(len(threads), server)
    thread.start()
    threads.append(thread)

  while not LorenzThread.stop_threads:
    try:
      LorenzThread.entrieslock.acquire()
      write_initials_file(initialfile)
      LorenzThread.entrieslock.release()
      time.sleep(5)
    except KeyboardInterrupt:
      LorenzThread.stop_threads = True
  print "Writing"
  print initialfile

  print "Joining Main Thread"
  for thread in threads:
    thread.join()
  print "Exiting Main Thread"

main()
