#!/usr/bin/python
import re
import sys
import os
import fileinput
from decimal import Decimal, getcontext

if (len(sys.argv) < 6):
  print "Error: pass x y z z_min z_max"
  sys.exit(-1)

pdim1 = int(sys.argv[1])
pdim2 = int(sys.argv[2])
pdim3 = int(sys.argv[3])
colormin = float(sys.argv[4])
colormax = float(sys.argv[5])

def floatify (sd):
  return float(sd.replace('~','-'))

def add(xs, ys):
  assert len(xs) == len(ys)
  return [x + y for x, y in zip(xs, ys)]

def uminus(xs):
  return [-x for x in xs]

def uminuss(xs):
  return [uminus(x) for x in xs]

def det3_pdevs(x, y):
  (x1, x2) = x
  (y1, y2) = y
  return x1 * y2 - x2 * y1

def cmp_pdevs(x, y):
  r = det3_pdevs(x, y)
  if r < 0:
    return -1
  elif r > 0:
    return 1
  else:
    return 0

def up_pdev(pdev):
  (x0, x1) = (pdev[pdim1], pdev[pdim2])
  if x1 == 0:
    if x0 < 0:
      return (-x0, 0)
    return (x0, x1)
  elif x1 < 0:
    return (-x0, -x1)
  else:
    return (x0, x1)

def polychain_of(c, pdevs):
  try:
    hd, tl = pdevs[0], pdevs[1:]
    n = add(c, hd)
    return [(c, n)] + polychain_of(n, tl)
  except IndexError:  # empty list: [][0] raises IndexError
    return []

def segments (c, pdevs):
  updevs = [up_pdev(pdev) for pdev in pdevs]
  updevs.sort(cmp_pdevs)
  dpdevs = [(2 * x, 2 * y) for (x, y) in updevs]
  for pdev in updevs:
    c = add(c, uminus(pdev))
  return (polychain_of(c, dpdevs + uminuss(dpdevs)))

def print_segment(seg):
  (x0, x1) = seg
  d = add(x1, uminus(x0))
  line = ''
  for x in x0:
    line = line + str(x) + ' '
  for x in d:
    line = line + str(x) + ' '
  print line

idx = 0
minx = 100000
maxx = -100000
miny = 100000
maxy = -100000

def intensity_of(x):
  i = max (0.0, min ((x - colormin) / (colormax - colormin), 1.0))
  return i

def hexcolor_of(x):
  i = intensity_of(x)
  M = 191
  m = 0
  c = int(m + (M - m) * i)
  h = str(hex(c))[2:]
  return h + h + h

def print_poly(idx, segs, color):
  global minx
  global maxx
  global miny
  global maxy
  line = 'set object ' + str(idx) + ' polygon from '
  ((x0, y0), (x1, y1)), tl = segs[0], segs[1:]
  line = line + str(x0) + ',' + str(y0)
  line = line + ' to '
  line = line + str(x1) + ',' + str(y1)
  minx = min(minx, x0, x1)
  miny = min(miny, y0, y1)
  maxx = max(maxx, x0, x1)
  maxy = max(maxy, y0, y1)
  for seg in tl:
    ((x0, y0), (x1, y1)) = seg
    minx = min(minx, x0, x1)
    miny = min(miny, y0, y1)
    maxx = max(maxx, x0, x1)
    maxy = max(maxy, y0, y1)
    line = line + ' to ' + str(x1) + ',' + str(y1)
  print line
  print 'set object ' + str(idx) + ' fc rgb "#' + hexcolor_of(color) + '" fillstyle solid 1.0 noborder'

while 1:
  try:
    line = sys.stdin.readline()
  except KeyboardInterrupt:
    break
  if not line:
    break
  if line.strip() == '':
    continue
  if not (line.startswith('aform(')):
    continue
  [dim, aform] = line.split('aform(', 1)[1].split('):', 1)
  dim = int(dim)
  aform = aform.replace('--','');
  aform = aform.replace('(','');
  aform = aform.replace(')','');
  aform = aform.replace(',','');
  pdevs = [floatify(sd) for sd in aform.split()];
  pdevs = [pdevs[n:n+dim] for n in range(0, len(pdevs), dim)]
  c = pdevs[0]
  color = c[pdim3]
  c = (c[pdim1], c[pdim2])
  pdevs = pdevs[1:]
  idx = idx + 1
  print_poly (idx, segments (c, pdevs), color)

print 'plot ' + '[' + str(minx) + ':' + str(maxx) +']  [' + str(miny) + ':' + str(maxy) + '] ' + 'NaN notitle'